import numpy as np
import pandas as pd
from unittest import TestCase  # or `from unittest import ...` if on Python 3.4+

from tests.helpers import verify_numeric


class TestHelpers(TestCase):

    def test_is_numeric_pandas(self):
        # Whole numbers, regardless of the byte length, should not raise AssertionError
        X = pd.DataFrame(np.ones([5, 5]), dtype='int32')
        verify_numeric(pd.DataFrame(X))

        X = pd.DataFrame(np.ones([5, 5]), dtype='int64')
        verify_numeric(pd.DataFrame(X))

        # Strings should raise AssertionError
        X = pd.DataFrame([['a', 'b', 'c'], ['d', 'e', 'f']])
        with self.assertRaises(Exception):
            verify_numeric(pd.DataFrame(X))

    def test_is_numeric_numpy(self):
        # Whole numbers, regardless of the byte length, should not raise AssertionError
        X = np.ones([5, 5], dtype='int32')
        verify_numeric(pd.DataFrame(X))

        X = np.ones([5, 5], dtype='int64')
        verify_numeric(pd.DataFrame(X))

        # Floats
        X = np.ones([5, 5], dtype='float32')
        verify_numeric(pd.DataFrame(X))

        X = np.ones([5, 5], dtype='float64')
        verify_numeric(pd.DataFrame(X))

    def test_verify_raises_AssertionError_on_categories(self):
        # Categories should raise AssertionError
        X = pd.DataFrame([['a', 'b', 'c'], ['d', 'e', 'f']], dtype='category')
        with self.assertRaises(Exception):
            verify_numeric(pd.DataFrame(X))



import pandas as pd
from unittest import TestCase  # or `from unittest import ...` if on Python 3.4+
import tests.helpers as th
import numpy as np
import category_encoders as encoders


np_X = th.create_array(n_rows=100)
np_X_t = th.create_array(n_rows=50, extras=True)
np_y = np.random.randn(np_X.shape[0]) > 0.5
np_y_t = np.random.randn(np_X_t.shape[0]) > 0.5
X = th.create_dataset(n_rows=100)
X_t = th.create_dataset(n_rows=50, extras=True)
y = pd.DataFrame(np_y)
y_t = pd.DataFrame(np_y_t)


class TestOrdinalEncoder(TestCase):
    def test_ordinal(self):

        enc = encoders.OrdinalEncoder(verbose=1, return_df=True)
        enc.fit(X)
        out = enc.transform(X_t)
        self.assertEqual(len(set(out["extra"].values)), 4)
        self.assertIn(-1, set(out["extra"].values))
        self.assertFalse(enc.mapping is None)
        self.assertTrue(len(enc.mapping) > 0)

        enc = encoders.OrdinalEncoder(verbose=1, mapping=enc.mapping, return_df=True)
        enc.fit(X)
        out = enc.transform(X_t)
        self.assertEqual(len(set(out["extra"].values)), 4)
        self.assertIn(-1, set(out["extra"].values))
        self.assertTrue(len(enc.mapping) > 0)

        enc = encoders.OrdinalEncoder(verbose=1, return_df=True, handle_unknown="return_nan")
        enc.fit(X)
        out = enc.transform(X_t)
        out_cats = [x for x in set(out["extra"].values) if np.isfinite(x)]
        self.assertEqual(len(out_cats), 3)
        self.assertFalse(enc.mapping is None)

    def test_ordinal_dist(self):
        data = np.array([["apple", "lemon"], ["peach", None]])
        encoder = encoders.OrdinalEncoder()
        result = encoder.fit_transform(data)
        self.assertEqual(2, len(result[0].unique()), "We expect two unique values in the column")
        self.assertEqual(2, len(result[1].unique()), "We expect two unique values in the column")
        self.assertFalse(np.isnan(result.values[1, 1]))

        encoder = encoders.OrdinalEncoder(handle_missing="return_nan")
        result = encoder.fit_transform(data)
        self.assertEqual(2, len(result[0].unique()), "We expect two unique values in the column")
        self.assertEqual(2, len(result[1].unique()), "We expect two unique values in the column")

    def test_pandas_categorical(self):
        X = pd.DataFrame(
            {
                "Str": ["a", "c", "c", "d"],
                "Categorical": pd.Categorical(list("bbea"), categories=["e", "a", "b"], ordered=True),
            }
        )

        enc = encoders.OrdinalEncoder()
        out = enc.fit_transform(X)

        th.verify_numeric(out)
        self.assertEqual(3, out["Categorical"][0])
        self.assertEqual(3, out["Categorical"][1])
        self.assertEqual(1, out["Categorical"][2])
        self.assertEqual(2, out["Categorical"][3])

    def test_handle_missing_have_nan_fit_time_expect_as_category(self):
        train = pd.DataFrame(
            {
                "city": ["chicago", np.nan],
                "city_cat": pd.Categorical(["chicago", np.nan]),
            }
        )

        enc = encoders.OrdinalEncoder(handle_missing="value")
        out = enc.fit_transform(train)

        self.assertListEqual([1, 2], out["city"].tolist())
        self.assertListEqual([1, 2], out["city_cat"].tolist())

    def test_handle_missing_have_nan_transform_time_expect_negative_2(self):
        train = pd.DataFrame(
            {
                "city": ["chicago", "st louis"],
                "city_cat": pd.Categorical(["chicago", "st louis"]),
            }
        )
        test = pd.DataFrame(
            {
                "city": ["chicago", np.nan],
                "city_cat": pd.Categorical(["chicago", np.nan]),
            }
        )

        enc = encoders.OrdinalEncoder(handle_missing="value")
        enc.fit(train)
        out = enc.transform(test)

        self.assertListEqual([1, -2], out["city"].tolist())
        self.assertListEqual([1, -2], out["city_cat"].tolist())

    def test_handle_unknown_have_new_value_expect_negative_1(self):
        # See #238
        train = pd.DataFrame({"city": ["chicago", "st louis"]})
        test = pd.DataFrame({"city": ["chicago", "los angeles"]})
        expected = [1.0, -1.0]

        enc = encoders.OrdinalEncoder(handle_missing="return_nan")
        enc.fit(train)
        result = enc.transform(test)["city"].tolist()

        self.assertEqual(expected, result)

    def test_handle_unknown_have_new_value_expect_negative_1_categorical(self):
        cities = ["st louis", "chicago", "los angeles"]
        train = pd.DataFrame({"city": pd.Categorical(cities[:-1], categories=cities)})
        test = pd.DataFrame({"city": pd.Categorical(cities[1:], categories=cities)})

        expected = [2.0, -1.0]

        enc = encoders.OrdinalEncoder(handle_missing="return_nan")
        enc.fit(train)
        result = enc.transform(test)["city"].tolist()

        self.assertEqual(expected, result)

    def test_custom_mapping(self):
        # See issue 193
        custom_mapping = [
            {
                "col": "col1",
                "mapping": {np.NaN: 0, "a": 1, "b": 2},
            },  # The mapping from the documentation
            {"col": "col2", "mapping": {np.NaN: -3, "x": 11, "y": 2}},
        ]
        custom_mapping_series = [
            {
                "col": "col1",
                "mapping": pd.Series({np.NaN: 0, "a": 1, "b": 2}),
            },  # The mapping from the documentation
            {"col": "col2", "mapping": pd.Series({np.NaN: -3, "x": 11, "y": 2})},
        ]

        train = pd.DataFrame({"col1": ["a", "a", "b", np.NaN], "col2": ["x", "y", np.NaN, np.NaN]})

        for mapping in [custom_mapping, custom_mapping_series]:
            with self.subTest():
                enc = encoders.OrdinalEncoder(handle_missing="value", mapping=mapping)
                out = enc.fit_transform(
                    train
                )  # We have to first 'fit' before 'transform' even if we do nothing during the fit...

                self.assertListEqual([1, 1, 2, 0], out["col1"].tolist())
                self.assertListEqual([11, 2, -3, -3], out["col2"].tolist())

    def test_HaveNegativeOneInTrain_ExpectCodedAsOne(self):
        train = pd.DataFrame({"city": [-1]})
        expected = [1]

        enc = encoders.OrdinalEncoder(cols=["city"])
        result = enc.fit_transform(train)["city"].tolist()

        self.assertEqual(expected, result)

    def test_HaveNaNInTrain_ExpectCodedAsOne(self):
        train = pd.DataFrame({"city": [np.nan]})
        expected = [1]

        enc = encoders.OrdinalEncoder(cols=["city"])
        result = enc.fit_transform(train)["city"].tolist()

        self.assertEqual(expected, result)

    def test_HaveNoneAndNan_ExpectCodesAsOne(self):
        train = pd.DataFrame({"city": [np.nan, None]})
        expected = [1, 1]

        enc = encoders.OrdinalEncoder(cols=["city"])
        result = enc.fit_transform(train)["city"].tolist()

        self.assertEqual(expected, result)

        new_nan = pd.DataFrame(
            {
                "city": [
                    np.nan,
                ]
            }
        )
        result_new_nan = enc.transform(new_nan)["city"].tolist()
        expected_new_nan = [1]
        self.assertEqual(expected_new_nan, result_new_nan)

        new_none = pd.DataFrame(
            {
                "city": [
                    None,
                ]
            }
        )
        result_new_none = enc.transform(new_none)["city"].tolist()
        expected_new_none = [1]
        self.assertEqual(expected_new_none, result_new_none)

    def test_inverse_transform_HaveUnknown_ExpectWarning(self):
        train = pd.DataFrame({"city": ["chicago", "st louis"]})
        test = pd.DataFrame({"city": ["chicago", "los angeles"]})

        enc = encoders.OrdinalEncoder(handle_missing="value", handle_unknown="value")
        enc.fit(train)
        result = enc.transform(test)

        message = (
            "inverse_transform is not supported because transform impute " "the unknown category -1 when encode city"
        )

        with self.assertWarns(UserWarning, msg=message) as w:
            enc.inverse_transform(result)

    def test_inverse_transform_HaveNanInTrainAndHandleMissingValue_ExpectReturnedWithNan(
        self,
    ):
        train = pd.DataFrame({"city": ["chicago", np.nan]})

        enc = encoders.OrdinalEncoder(handle_missing="value", handle_unknown="value")
        result = enc.fit_transform(train)
        original = enc.inverse_transform(result)

        pd.testing.assert_frame_equal(train, original)

    def test_inverse_transform_HaveNanInTrainAndHandleMissingReturnNan_ExpectReturnedWithNan(
        self,
    ):
        train = pd.DataFrame({"city": ["chicago", np.nan]})

        enc = encoders.OrdinalEncoder(handle_missing="return_nan", handle_unknown="value")
        result = enc.fit_transform(train)
        original = enc.inverse_transform(result)

        pd.testing.assert_frame_equal(train, original)

    def test_inverse_transform_BothFieldsAreReturnNanWithNan_ExpectValueError(self):
        train = pd.DataFrame({"city": ["chicago", np.nan]})
        test = pd.DataFrame({"city": ["chicago", "los angeles"]})

        enc = encoders.OrdinalEncoder(handle_missing="return_nan", handle_unknown="return_nan")
        enc.fit(train)
        result = enc.transform(test)

        message = (
            "inverse_transform is not supported because transform impute " "the unknown category nan when encode city"
        )

        with self.assertWarns(UserWarning, msg=message) as w:
            enc.inverse_transform(result)

    def test_inverse_transform_HaveMissingAndNoUnknown_ExpectInversed(self):
        train = pd.DataFrame({"city": ["chicago", np.nan]})
        test = pd.DataFrame({"city": ["chicago", "los angeles"]})

        enc = encoders.OrdinalEncoder(handle_missing="value", handle_unknown="return_nan")
        enc.fit(train)
        result = enc.transform(test)
        original = enc.inverse_transform(result)

        pd.testing.assert_frame_equal(train, original)

    def test_inverse_transform_HaveHandleMissingValueAndHandleUnknownReturnNan_ExpectBestInverse(
        self,
    ):
        train = pd.DataFrame({"city": ["chicago", np.nan]})
        test = pd.DataFrame({"city": ["chicago", np.nan, "los angeles"]})
        expected = pd.DataFrame({"city": ["chicago", np.nan, np.nan]})

        enc = encoders.OrdinalEncoder(handle_missing="value", handle_unknown="return_nan")
        enc.fit(train)
        result = enc.transform(test)
        original = enc.inverse_transform(result)

        pd.testing.assert_frame_equal(expected, original)

    def test_inverse_with_mapping(self):
        df = X.copy(deep=True)
        categoricals = [
            "unique_int",
            "unique_str",
            "invariant",
            "underscore",
            "none",
            "extra",
        ]
        mappings = {
            "as Series": [
                {
                    "col": c,
                    "mapping": pd.Series(data=range(len(df[c].unique())), index=df[c].unique()),
                    "data_type": X[c].dtype,
                }
                for c in categoricals
            ],
            "as Dict": [{"col": c, "mapping": {k: idx for idx, k in enumerate(df[c].unique())}} for c in categoricals],
        }
        for msg, mapping in mappings.items():
            with self.subTest(msg):
                df = X.copy(deep=True)
                enc = encoders.OrdinalEncoder(
                    cols=categoricals,
                    handle_unknown="ignore",
                    mapping=mapping,
                    return_df=True,
                )
                df[categoricals] = enc.fit_transform(df[categoricals])
                recovered = enc.inverse_transform(df[categoricals])
                pd.testing.assert_frame_equal(X[categoricals], recovered)

    def test_validate_mapping(self):
        custom_mapping = [
            {
                "col": "col1",
                "mapping": {np.NaN: 0, "a": 1, "b": 2},
            },  # The mapping from the documentation
            {"col": "col2", "mapping": {np.NaN: -3, "x": 11, "y": 2}},
        ]
        expected_valid_mapping = [
            {
                "col": "col1",
                "mapping": pd.Series({np.NaN: 0, "a": 1, "b": 2}),
            },  # The mapping from the documentation
            {"col": "col2", "mapping": pd.Series({np.NaN: -3, "x": 11, "y": 2})},
        ]
        enc = encoders.OrdinalEncoder()
        actual_valid_mapping = enc._validate_supplied_mapping(custom_mapping)
        self.assertEqual(len(actual_valid_mapping), len(expected_valid_mapping))
        for idx in range(len(actual_valid_mapping)):
            self.assertEqual(actual_valid_mapping[idx]["col"], expected_valid_mapping[idx]["col"])
            pd.testing.assert_series_equal(actual_valid_mapping[idx]["mapping"], expected_valid_mapping[idx]["mapping"])



import numpy as np
import pandas as pd
from category_encoders import OrdinalEncoder
import category_encoders.utils as util


class RankHotEncoder(util.BaseEncoder, util.UnsupervisedTransformerMixin):
    """The rank-hot encoder is similar to a one-hot encoder,
    except every feature up to and including the current rank is hot.
    This is also called thermometer encoding.

    Parameters
    ----------

    verbose: int
        integer indicating verbosity of the output. 0 for none.
    cols: list
        a list of columns to encode, if None, all string columns will be encoded.
    drop_invariant: bool
        boolean for whether or not to drop columns with 0 variance.
    use_cat_names: bool
        if True, category values will be included in the encoded column names.
           Since this can result in duplicate column names,
           duplicates are suffixed with '#' symbol until a unique name is generated.
        If False, category indices will be used instead of the category values.
    handle_unknown: str
        options are 'error', 'value', 'return_nan'.
        The default is 'value'.
        'value': If an unknown label occurrs, it is represented as 0 array.
        'error': If an unknown label occurrs, error message is displayed.
        'return_nan': If an unknown label occurrs, np.nan is returned in all columns.
    handle_missing: str
        options are 'error', 'value' and 'return_nan'. The default is 'value'.
        Missing value also considered as unknown value in the final data set.

    Example
    -------
     >>> from category_encoders import *
     >>> import pandas as pd
     >>> from sklearn.datasets import fetch_openml
     >>> bunch = fetch_openml(name="house_prices", as_frame=True)
     >>> display_cols = ["Id", "MSSubClass", "MSZoning", "LotFrontage", "YearBuilt", "Heating", "CentralAir"]
     >>> y = bunch.target
     >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)[display_cols]
     >>> enc = RankHotEncoder(cols=['CentralAir', 'Heating'], handle_unknown='indicator').fit(X, y)
     >>> numeric_dataset = enc.transform(X)
     >>> print(numeric_dataset.info())
     <class 'pandas.core.frame.DataFrame'>
     RangeIndex: 1460 entries, 0 to 1459
     Data columns (total 13 columns):
      #   Column        Non-Null Count  Dtype  
     ---  ------        --------------  -----  
      0   Id            1460 non-null   float64
      1   MSSubClass    1460 non-null   float64
      2   MSZoning      1460 non-null   object 
      3   LotFrontage   1201 non-null   float64
      4   YearBuilt     1460 non-null   float64
      5   Heating_1     1460 non-null   int64  
      6   Heating_2     1460 non-null   int64  
      7   Heating_3     1460 non-null   int64  
      8   Heating_4     1460 non-null   int64  
      9   Heating_5     1460 non-null   int64  
      10  Heating_6     1460 non-null   int64  
      11  CentralAir_1  1460 non-null   int64  
      12  CentralAir_2  1460 non-null   int64  
     dtypes: float64(4), int64(8), object(1)
     memory usage: 148.4+ KB
     None
    """

    prefit_ordinal = True
    encoding_relation = util.EncodingRelation.ONE_TO_N_UNIQUE

    def __init__(
        self,
        verbose=0,
        cols=None,
        drop_invariant=False,
        return_df=True,
        handle_missing="value",
        handle_unknown="value",
        use_cat_names=None,
    ):
        super().__init__(
            verbose=verbose,
            cols=cols,
            drop_invariant=drop_invariant,
            return_df=return_df,
            handle_unknown=handle_unknown,
            handle_missing=handle_missing,
        )
        self._dim = None
        self.mapping = None
        self.use_cat_names = use_cat_names

    def _fit(self, X, y, **kwargs):
        oe_missing_strat = {
            'error': 'error',
            'return_nan': 'return_nan',
            'value': 'value',
            'indicator': 'return_nan',
        }[self.handle_missing]
        # supply custom mapping in order to assure order of ordinal variable
        ordered_mapping = []
        for col in self.cols:
            oe_col = OrdinalEncoder(verbose=self.verbose, cols=[col], handle_unknown="value", handle_missing=oe_missing_strat)
            oe_col.fit(X[col].sort_values().to_frame(name=col))
            ordered_mapping += oe_col.mapping

        self.ordinal_encoder = OrdinalEncoder(
            verbose=self.verbose, cols=self.cols, handle_unknown="value", handle_missing=oe_missing_strat, mapping=ordered_mapping
        )
        self.ordinal_encoder = self.ordinal_encoder.fit(X)

        self.mapping = self.generate_mapping()

        return self

    def _transform(self, X_in, override_return_df=False):
        X = X_in.copy(deep=True)
        X = self.ordinal_encoder.transform(X)
        input_cols = X.columns.values.tolist()

        if self.handle_unknown == "error":
            if X[self.cols].isin([-1]).any().any():
                raise ValueError("Columns to be encoded can not contain new values")

        for switch, ordinal_switch in zip(self.mapping, self.ordinal_encoder.category_mapping):
            col = switch.get("col")
            mod = switch.get("mapping")
            encode_feature_series = X[col]

            unknow_elements = encode_feature_series[encode_feature_series == -1]

            encoding_dict = {i: list(row.values()) for i, row in mod.to_dict(orient="index").items()}
            if self.handle_unknown == "value":
                default_value = [0] * len(encoding_dict)
            elif self.handle_unknown == "return_nan":
                default_value = [np.nan] * len(encoding_dict)
            elif self.handle_unknown == "error":
                if not unknow_elements.empty:
                    unknowns_str = ', '.join([str(x) for x in unknow_elements.unique()])
                    msg = f"Unseen values {unknowns_str} during transform in column {col}."
                    raise ValueError(msg)
                default_value = [0] * len(encoding_dict)
            else:
                raise ValueError(f"invalid option for 'handle_unknown' parameter: {self.handle_unknown}")

            def apply_coding(row: pd.Series):
                val = row.iloc[0]
                if pd.isna(val):
                    if self.handle_missing == "value":
                        return default_value
                    elif self.handle_missing == "return_nan":
                        return [np.nan] * len(default_value)
                    else:
                        raise ValueError("Unhandled NaN")
                return encoding_dict.get(row.iloc[0], default_value)

            encoded = encode_feature_series.to_frame().apply(apply_coding, axis=1, result_type="expand")
            encoded.columns = mod.columns

            X = pd.concat([encoded, X], axis=1)

            old_column_index = input_cols.index(col)
            input_cols[old_column_index:old_column_index + 1] = mod.columns
        X = X.reindex(columns=input_cols)

        return X

    def create_dataframe(self, X, encoded, key_col):

        if not (isinstance(encoded, pd.DataFrame) or isinstance(encoded, pd.Series)):
            encoded = pd.DataFrame(encoded, columns=key_col)

        X_ = pd.concat([encoded, X], axis=1)
        return X_

    def inverse_transform(self, X_in):
        X = X_in.copy(deep=True)
        cols = X.columns.values.tolist()
        if self._dim is None:
            raise ValueError("Must train encoder before it can be used to inverse_transform data")

        for switch, ordinal_mapping in zip(self.mapping, self.ordinal_encoder.category_mapping):
            col = switch.get("col")
            cats = switch.get("mapping")
            if col != ordinal_mapping.get("col"):
                raise ValueError("Column order of OrdinalEncoder and RankHotEncoder do not match")
            inv_map = {v: k for k, v in ordinal_mapping.get("mapping").to_dict().items()}

            arrs = X[cats.columns]
            reencode = arrs.sum(axis=1).rename(col)

            orig_dtype = ordinal_mapping.get("data_type")
            reencode2 = reencode.replace(inv_map).astype(orig_dtype)
            if np.any(reencode2[:] == 0):
                reencode2[reencode2[:] == 0] = np.nan

            X = self.create_dataframe(X, reencode2, col)

            first_inex = cols.index(cats.columns[0])
            last_index = cols.index(cats.columns[-1]) + 1

            del cols[first_inex:last_index]
            cols.insert(self.ordinal_encoder.feature_names_out_.index(col), col)

        X = X.reindex(columns=cols)

        return X

    def generate_mapping(self):
        mapping = []
        found_column_counts = {}

        for switch in self.ordinal_encoder.mapping:
            col: str = switch.get("col")
            values: pd.Series = switch.get("mapping").copy(deep=True)

            if self.handle_missing == "value":
                values = values[values > 0]

            if len(values) == 0:
                continue

            index = []
            new_columns = []

            for cat_name, class_ in values.iteritems():
                if self.use_cat_names:
                    n_col_name = f"{col}_{cat_name}"
                    found_count = found_column_counts.get(n_col_name, 0)
                    found_column_counts[n_col_name] = found_count + 1
                    n_col_name += "#" * found_count
                else:
                    n_col_name = f"{col}_{class_}"

                index.append(class_)
                new_columns.append(n_col_name)

            base_matrix = np.tril(np.ones((len(index), len(index)), dtype=int))
            base_df = pd.DataFrame(data=base_matrix, columns=new_columns, index=index)

            mapping.append({"col": col, "mapping": base_df})
        return mapping
